class WrongOpenAIClassException(Exception):
    pass


class InvalidSamplingRatioException(Exception):
    pass


class InvalidMonaCredsException(Exception):
    pass


class InvalidLagnchainLLMException(Exception):
    pass
